@@ -8,7 +8,7 @@ use FFI::Platypus::Function;
 use FFI::Platypus::Type;
 
 # ABSTRACT: Write Perl bindings to non-Perl libraries with FFI. No XS required.
-our $VERSION = '1.10'; # VERSION
+our $VERSION = '1.31'; # VERSION
 
 # Platypus Man,
 # Platypus Man,
@@ -72,6 +72,10 @@ sub new
   {
     $tp = 'Version1';
   }
+  elsif($api == 2)
+  {
+    $tp = 'Version1';
+  }
   else
   {
     Carp::croak("API version $api not (yet) implemented");
@@ -168,12 +172,17 @@ sub lang
       $type_map{pointer} = 'opaque' if $self->{tp}->isa('FFI::Platypus::TypeParser::Version0');
       $self->{tp}->type_map(\%type_map);
     }
+
+    $class->load_custom_types($self) if $class->can('load_custom_types');
   }
 
   $self->{lang};
 }
 
 
+sub api { shift->{api} }
+
+
 sub type
 {
   my($self, $name, $alias) = @_;
@@ -276,7 +285,7 @@ sub function
   my $wrapper;
   $wrapper = pop if ref $_[-1] eq 'CODE';
 
-  croak "usage \$ffi->function( name, [ arguments ], return_type)" unless @_ >= 4 && @_ <= 6;
+  croak "usage \$ffi->function( \$name, \\\@arguments, [\\\@var_args], [\$return_type])" unless @_ >= 3 && @_ <= 6;
 
   my $self = shift;
   my $name = shift;
@@ -284,6 +293,7 @@ sub function
   my $var_args;
   $var_args = shift if defined $_[0] && ref($_[0]) eq 'ARRAY';
   my $ret = shift;
+  $ret = 'void' unless defined $ret;
 
   # special case: treat a single void argument type as an empty list of
   # arguments, a la olde timey C compilers.
@@ -373,10 +383,17 @@ sub cast
 
 sub attach_cast
 {
-  my($self, $name, $type1, $type2) = @_;
+  my($self, $name, $type1, $type2, $wrapper) = @_;
   my $caller = caller;
   $name = join '::', $caller, $name unless $name =~ /::/;
-  $self->attach([0 => $name] => [$type1] => $type2 => '$');
+  if(defined $wrapper && ref($wrapper) eq 'CODE')
+  {
+    $self->attach([0 => $name] => [$type1] => $type2 => '$', $wrapper);
+  }
+  else
+  {
+    $self->attach([0 => $name] => [$type1] => $type2 => '$');
+  }
   $self;
 }
 
@@ -399,6 +416,47 @@ sub alignof
 }
 
 
+sub kindof
+{
+  my($self, $name) = @_;
+  ref $self
+    ? $self->{tp}->parse($name)->kindof
+    : $self->new->kindof($name);
+}
+
+
+sub countof
+{
+  my($self, $name) = @_;
+  ref $self
+    ? $self->{tp}->parse($name)->countof
+    : $self->new->countof($name);
+}
+
+
+sub def
+{
+  my $self = shift;
+  my $package = shift || caller;
+  my $type = shift;
+  if(@_)
+  {
+    $self->type($type);
+    $self->{def}->{$package}->{$type} = shift;
+  }
+  $self->{def}->{$package}->{$type};
+}
+
+
+sub unitof
+{
+  my($self, $name) = @_;
+  ref $self
+    ? $self->{tp}->parse($name)->unitof
+    : $self->new->unitof($name);
+}
+
+
 sub find_lib
 {
   my $self = shift;
@@ -525,7 +583,7 @@ FFI::Platypus - Write Perl bindings to non-Perl libraries with FFI. No XS requir
 
 =head1 VERSION
 
-version 1.10
+version 1.31
 
 =head1 SYNOPSIS
 
@@ -546,6 +604,7 @@ version 1.10
 
 Platypus is a library for creating interfaces to machine code libraries
 written in languages like C, L<C++|FFI::Platypus::Lang::CPP>,
+L<Go|FFI::Platypus::Lang::Go>,
 L<Fortran|FFI::Platypus::Lang::Fortran>,
 L<Rust|FFI::Platypus::Lang::Rust>,
 L<Pascal|FFI::Platypus::Lang::Pascal>. Essentially anything that gets
@@ -599,6 +658,7 @@ least some understanding of C, the C pre-processor, and some C++ caveats
 (since on some platforms Perl is compiled and linked with a C++
 compiler). Platypus on the other hand could be used to call other
 compiled languages, like L<Fortran|FFI::Platypus::Lang::Fortran>,
+L<Go|FFI::Platypus::Lang::Go>,
 L<Rust|FFI::Platypus::Lang::Rust>,
 L<Pascal|FFI::Platypus::Lang::Pascal>, L<C++|FFI::Platypus::Lang::CPP>,
 or even L<assembly|FFI::Platypus::Lang::ASM>, allowing you to focus
@@ -767,6 +827,14 @@ Platypus to find symbols using the demangled names (for example, if you
 specify L<CPP|FFI::Platypus::Lang::CPP> for C++ you can use method names
 like C<Foo::get_bar()> with L</attach> or L</function>.
 
+=head2 api
+
+[version 1.11]
+
+ my $level = $ffi->api;
+
+Returns the API level of the Platypus instance.
+
 =head1 METHODS
 
 =head2 type
@@ -912,6 +980,8 @@ Examples:
 
  my $function = $ffi->function( $name => \@fixed_argument_types => \@var_argument_types => $return_type);
  my $function = $ffi->function( $name => \@fixed_argument_types => \@var_argument_types => $return_type, \&wrapper);
+ my $function = $ffi->function( $name => \@fixed_argument_types => \@var_argument_types);
+ my $function = $ffi->function( $name => \@fixed_argument_types => \@var_argument_types => \&wrapper);
 
 Version 0.91 and later allows you to creat functions for c variadic functions
 (such as printf, scanf, etc) which can take a variable number of arguments.
@@ -931,6 +1001,10 @@ Some older versions of libffi and possibly some platforms may not support
 variadic functions.  If you try to create a one, then an exception will be
 thrown.
 
+[version 1.26]
+
+If the return type is omitted then C<void> will be the assumed return type.
+
 =head2 attach
 
  $ffi->attach($name => \@argument_types => $return_type);
@@ -1015,11 +1089,17 @@ Something that won't work is trying to cast an array to anything:
 =head2 attach_cast
 
  $ffi->attach_cast("cast_name", $original_type, $converted_type);
+ $ffi->attach_cast("cast_name", $original_type, $converted_type, \&wrapper);
  my $converted_value = cast_name($original_value);
 
 This function attaches a cast as a permanent xsub.  This will make it
 faster and may be useful if you are calling a particular cast a lot.
 
+[version 1.26]
+
+A wrapper may be added as the last argument to C<attach_cast> and works
+just like the wrapper for C<attach> and C<function> methods.
+
 =head2 sizeof
 
  my $size = $ffi->sizeof($type);
@@ -1053,6 +1133,65 @@ iterations.
 
 Returns the alignment of the given type in bytes.
 
+=head2 kindof
+
+[version 1.24]
+
+ my $kind = $ffi->kindof($type);
+
+Returns the kind of a type.  This is a string with a value of one of
+
+=over 4
+
+=item C<void>
+
+=item C<scalar>
+
+=item C<string>
+
+=item C<closure>
+
+=item C<record>
+
+=item C<record-value>
+
+=item C<pointer>
+
+=item C<array>
+
+=item C<object>
+
+=back
+
+=head2 countof
+
+[version 1.24]
+
+ my $count = $ffi->countof($type);
+
+For array types returns the number of elements in the array (returns 0 for variable length array).
+For the C<void> type returns 0.  Returns 1 for all other types.
+
+=head2 def
+
+[version 1.24]
+
+ $ffi->def($package, $type, $value);
+ my $value = $ff->def($package, $type);
+
+This method allows you to store data for types.  If the C<$package> is not provided, then the
+caller's package will be used.  C<$type> must be a legal Platypus type for the L<FFI::Platypus>
+instance.
+
+=head2 unitof
+
+[version 1.24]
+
+ my $unittype = $ffi->unitof($type);
+
+For array and pointer types, returns the basic type without the array or pointer part.
+In other words, for C<sin16[]> or C<sint16*> it will return C<sint16>.
+
 =head2 find_lib
 
 [version 0.20]
@@ -1224,40 +1363,49 @@ L<FFI::Platypus::Memory> module.
 
 =head2 structured data records
 
- package My::UnixTime;
+ use FFI::Platypus 1.00;
+ use FFI::C;
  
- use FFI::Platypus::Record;
+ my $ffi = FFI::Platypus->new(
+   api => 1,
+   lib => [undef],
+ );
+ FFI::C->ffi($ffi);
  
- record_layout_1(qw(
-     int    tm_sec
-     int    tm_min
-     int    tm_hour
-     int    tm_mday
-     int    tm_mon
-     int    tm_year
-     int    tm_wday
-     int    tm_yday
-     int    tm_isdst
-     long   tm_gmtoff
-     string tm_zone
- ));
+ package Unix::TimeStruct {
  
- my $ffi = FFI::Platypus->new( api => 1 );
- $ffi->lib(undef);
- # define a record class My::UnixTime and alias it to "tm"
- $ffi->type("record(My::UnixTime)*" => 'tm');
+   FFI::C->struct(tm => [
+     tm_sec    => 'int',
+     tm_min    => 'int',
+     tm_hour   => 'int',
+     tm_mday   => 'int',
+     tm_mon    => 'int',
+     tm_year   => 'int',
+     tm_wday   => 'int',
+     tm_yday   => 'int',
+     tm_isdst  => 'int',
+     tm_gmtoff => 'long',
+     _tm_zone  => 'opaque',
+   ]);
  
- # attach the C localtime function as a constructor
- $ffi->attach( localtime => ['time_t*'] => 'tm', sub {
-   my($inner, $class, $time) = @_;
-   $time = time unless defined $time;
-   $inner->(\$time);
- });
+   # For now 'string' is unsupported by FFI::C, but we
+   # can cast the time zone from an opaque pointer to
+   # string.
+   sub tm_zone {
+     my $self = shift;
+     $ffi->cast('opaque', 'string', $self->_tm_zone);
+   }
  
- package main;
+   # attach the C localtime function
+   $ffi->attach( localtime => ['time_t*'] => 'tm', sub {
+     my($inner, $class, $time) = @_;
+     $time = time unless defined $time;
+     $inner->(\$time);
+   });
+ }
  
  # now we can actually use our My::UnixTime class
- my $time = My::UnixTime->localtime;
+ my $time = Unix::TimeStruct->localtime;
  printf "time is %d:%d:%d %s\n",
    $time->tm_hour,
    $time->tm_min,
@@ -1268,16 +1416,19 @@ B<Discussion>: C and other machine code languages frequently provide
 interfaces that include structured data records (known as "structs" in
 C).  They sometimes provide an API in which you are expected to
 manipulate these records before and/or after passing them along to C
-functions.  There are a few ways of dealing with such interfaces, but
-the easiest way is demonstrated here defines a record class using a
-specific layout.  For more details see L<FFI::Platypus::Record>.
-(L<FFI::Platypus::Type> includes some other ways of manipulating
-structured data records).
+functions.  For C pointers to structs, unions and arrays of structs and
+unions, the easiest interface to use is via L<FFI::C>.  If you are
+working with structs that must be passed as values (not pointers),
+then you want to use the L<FFI::Platypus::Record> class instead.
+We will discuss this class later.
 
-The C C<localtime> function takes a pointer to a record, hence we suffix
-the type with a star: C<record(My::UnixTime)*>.  If the function takes
-a record in pass-by-value mode then we'd just say C<record(My::UnixTime)>
-with no star suffix.
+The C C<localtime> function takes a pointer to a C struct.  We simply define
+the members of the struct using the L<FFI::C> C<struct> method.  Because
+we used the C<ffi> method to tell L<FFI::C> to use our local instance of
+L<FFI::Platypus> it registers the C<tm> type for us, and we can just start
+using it as a return type!
+
+=head2 structured data records by-value
 
 =head2 libuuid
 
@@ -1934,7 +2085,7 @@ That means all the way back to 5.8.1.  Unfortunately, 5.10.0 seems to have a pro
 that is difficult to diagnose.  Patches to fix are welcome, if you want to help
 out on this, please see:
 
-L<https://github.com/Perl5-FFI/FFI-Platypus/issues/68>
+L<https://github.com/PerlFFI/FFI-Platypus/issues/68>
 
 Since this is an older buggy version of Perl it is recommended that you instead
 upgrade to 5.10.1 or later.
@@ -1988,14 +2139,14 @@ If something does not work the way you think it should, or if you have a
 feature request, please open an issue on this project's GitHub Issue
 tracker:
 
-L<https://github.com/perl5-FFI/FFI-Platypus/issues>
+L<https://github.com/perlFFI/FFI-Platypus/issues>
 
 =head1 CONTRIBUTING
 
 If you have implemented a new feature or fixed a bug then you may make a
 pull request on this project's GitHub repository:
 
-L<https://github.com/Perl5-FFI/FFI-Platypus/pulls>
+L<https://github.com/PerlFFI/FFI-Platypus/pulls>
 
 This project is developed using L<Dist::Zilla>.  The project's git
 repository also comes with the C<Makefile.PL> file necessary
@@ -2119,7 +2270,7 @@ making significant changes to the Platypus Core.  For that I use
 
 =over 4
 
-=item L<https://github.com/Perl5-FFI/FFI-Performance>
+=item L<https://github.com/PerlFFI/FFI-Performance>
 
 =back
 
@@ -2149,6 +2300,14 @@ Type definitions for Platypus.
 Define structured data records (C "structs") for use with
 Platypus.
 
+=item L<FFI::C>
+
+Another interface for defining structured data records for use
+with Platypus.  Its advantage over L<FFI::Platypus::Record> is
+that it supports C<union>s and nested data structures.  Its
+disadvantage is that it doesn't support passing C<struct>s
+by-value.
+
 =item L<FFI::Platypus::API>
 
 The custom types API for Platypus.
@@ -2179,6 +2338,10 @@ language
 
 Documentation and tools for using Platypus with Fortran
 
+=item L<FFI::Platypus::Lang::Go>
+
+Documentation and tools for using Platypus with Go
+
 =item L<FFI::Platypus::Lang::Pascal>
 
 Documentation and tools for using Platypus with Free Pascal
@@ -2219,9 +2382,9 @@ recent versions of Perl.
 
 =item L<FFI>
 
-Foreign function interface based on (nomenclature is everything) FSF's
-C<ffcall>. It hasn't worked for quite some time, and C<ffcall> is no
-longer supported or distributed.
+Older, simpler, less featureful FFI.  It used to be implemented
+using FSF's C<ffcall>.  Because C<ffcall> has been unsupported for
+some time, I reimplemented this module using L<FFI::Platypus>.
 
 =item L<C::DynaLib>
 
@@ -2294,7 +2457,7 @@ Diab Jerius (DJERIUS)
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2015,2016,2017,2018,2019 by Graham Ollis.
+This software is copyright (c) 2015,2016,2017,2018,2019,2020 by Graham Ollis.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
